@@ -5,7 +5,7 @@ use warnings;
 use Socket;
 use Carp;
 
-our $VERSION = '0.03_03';
+our $VERSION = '0.14_01';
 
 =head1 NAME
 
@@ -58,13 +58,35 @@ until you call C<-E<gt>run()>.
 
 =cut
 
+# Handle SIGHUP
+
+local $SIG{CHLD} = 'IGNORE'; # reap child processes
+local $SIG{HUP} = sub {
+    # on a "kill -HUP", we first close our socket handles.
+    close Remote;
+    close HTTPDaemon;
+
+    # and then, on systems implementing fork(), we make sure
+    # we are running with a new pid, so another -HUP will still
+    # work on the new process.
+    require Config;
+    if ($Config::Config{d_fork} and my $pid = fork()) {
+        # finally, allow ^C on the parent process to terminate
+        # the children.
+        waitpid($pid, 0); exit;
+    }
+
+    # do the exec. if $0 is not executable, try running it with $^X.
+    exec { $0 } ( ((-x $0) ? () : ($^X)), $0, @ARGV );
+};
+
+
 
 sub new {
     my ($proto,$port) = @_;
     my $class = ref($proto) || $proto;
 
     if ( $class eq __PACKAGE__ ) {
-	warn "HTTP::Server::Simple called directly - using CGI version";
 	require HTTP::Server::Simple::CGI;
 	return HTTP::Server::Simple::CGI->new(@_[1..$#_]);
     }
@@ -134,79 +156,177 @@ start listening for http requests.
 
 =cut
 
+my $server_class_id = 0;
 sub run {
+    my $self    = shift;
+    my $server  = $self->net_server;
+
+    # $pkg is generated anew for each invocation to "run"
+    # Just so we can use different net_server() implementations
+    # in different runs.
+    my $pkg     = join '::', ref($self), "NetServer".$server_class_id++;
+
+    no strict 'refs';
+    *{"$pkg\::process_request"} = $self->_process_request;
+
+    if ($server) {
+        require join('/', split /::/, $server).'.pm';
+        *{"$pkg\::ISA"} = [$server];
+        $self->print_banner;
+    }
+    else {
+        $self->setup_listener;
+        *{"$pkg\::run"} = $self->_default_run;
+    }
+
+    $pkg->run(port => $self->port);
+}
+
+=head2 net_server
+
+User-overridable method. If you set it to a C<Net::Server> subclass,
+that subclass is used for the C<run> method.  Otherwise, a minimal 
+implementation is used as default.
+
+=cut
+
+sub net_server { undef }
+
+sub _default_run {
     my $self = shift;
 
-    $self->setup_listener;
+    # Default "run" closure method for a stub, minimal Net::Server instance.
+    sub {
+        my $pkg = shift;
 
-    $self->print_banner;
+        $self->print_banner;
 
-    while (1) {
+        while (1) {
+            for ( ; accept( Remote, HTTPDaemon ) ; close Remote ) {
+                $self->stdio_handle(\*Remote);
+                $self->accept_hook if $self->can("accept_hook");
 
-        for ( ; accept( Remote, HTTPDaemon ) ; close Remote ) {
+                *STDIN  = $self->stdin_handle();
+                *STDOUT = $self->stdout_handle();
+		select STDOUT; # required for Recorder
+                $pkg->process_request;
+            }
+        }    
+    }
+}
 
-	    $self->accept_hook if $self->can("accept_hook");
+sub _process_request {
+    my $self = shift;
 
-            *STDIN  = *Remote;
-            *STDOUT = *Remote;
+    # Create a callback closure that is invoked for each incoming request;
+    # the $self above is bound into the closure.
+    sub {
+        $self->stdio_handle(*STDIN) unless $self->stdio_handle;
 
-            my $remote_sockaddr = getpeername(STDIN);
-            my ( undef, $iaddr ) = sockaddr_in($remote_sockaddr);
-            my $peername = gethostbyaddr( $iaddr, AF_INET ) || "localhost";
+        # Default to unencoded, raw data out.
+        # if you're sending utf8 and latin1 data mixed, you may need to override this
+        binmode STDIN, ':raw';
+        binmode STDOUT, ':raw';
 
-            my $peeraddr = inet_ntoa($iaddr) || "127.0.0.1";
+        my $remote_sockaddr = getpeername($self->stdio_handle);
+        my ( undef, $iaddr ) = sockaddr_in($remote_sockaddr);
+        my $peername = gethostbyaddr( $iaddr, AF_INET ) || "localhost";
 
-            my $local_sockaddr = getsockname(STDIN);
-            my ( undef, $localiaddr ) = sockaddr_in($local_sockaddr);
-            my $localname = gethostbyaddr( $localiaddr, AF_INET )
-              || "localhost";
-            my $localaddr = inet_ntoa($localiaddr) || "127.0.0.1";
+        my $peeraddr = inet_ntoa($iaddr) || "127.0.0.1";
 
-            my ( $method, $request_uri, $proto ) =
-		$self->parse_request
-		    or do {$self->bad_request; next};
+        my $local_sockaddr = getsockname($self->stdio_handle);
+        my ( undef, $localiaddr ) = sockaddr_in($local_sockaddr);
+        my $localname = gethostbyaddr( $localiaddr, AF_INET )
+            || "localhost";
+        my $localaddr = inet_ntoa($localiaddr) || "127.0.0.1";
 
-	    $proto ||= "HTTP/0.9";
+        my ( $method, $request_uri, $proto ) =
+            $self->parse_request
+                or do {$self->bad_request; return};
 
-            my ( $file, $query_string ) =
-              ( $request_uri =~ /([^?]*)(?:\?(.*))?/ );    # split at ?
+        $proto ||= "HTTP/0.9";
 
-            $self->bad_request, next
-		if ( $method !~ /^(GET|POST|HEAD)$/ );
+        my ( $file, $query_string ) =
+            ( $request_uri =~ /([^?]*)(?:\?(.*))?/ );    # split at ?
 
-            $self->setup(
-                method       => $method,
-                protocol     => $proto,
-                query_string => ( $query_string || '' ),
-                request_uri  => $request_uri,
-                path         => $file,
-                localname    => $localname,
-                localport    => $self->port,
-                peername     => $peername,
-                peeraddr     => $peeraddr,
-            );
+        if ( $method !~ /^(?:GET|POST|HEAD)$/ ) {
+            $self->bad_request;
+            return;
+        }
 
-	    # HTTP/0.9 didn't have any headers (I think)
-	    if ( $proto =~ m{HTTP/(\d(\.\d)?)$} and $1 >= 1 ) {
+        $self->setup(
+            method       => $method,
+            protocol     => $proto,
+            query_string => ( $query_string || '' ),
+            request_uri  => $request_uri,
+            path         => $file,
+            localname    => $localname,
+            localport    => $self->port,
+            peername     => $peername,
+            peeraddr     => $peeraddr,
+        );
 
-		my $headers = $self->parse_headers
-		    or do{$self->bad_request; next};
+        # HTTP/0.9 didn't have any headers (I think)
+        if ( $proto =~ m{HTTP/(\d(\.\d)?)$} and $1 >= 1 ) {
 
-		$self->setup( headers => $headers
-			    ) if $headers;
+            my $headers = $self->parse_headers
+                or do{$self->bad_request; return};
 
-	    }
-
-	    $self->post_setup_hook if $self->can("post_setup_hook");
-
-            $self->handler;
+            $self->headers( $headers) ;
 
         }
 
-    }
+        $self->post_setup_hook if $self->can("post_setup_hook");
 
+        $self->handler;
+    }
 }
 
+
+=head2 stdio_handle [FILEHANDLE]
+
+When called with an argument, sets the socket to the server to that arg.
+
+Returns the socket to the server; you should only use this for actual socket-related
+calls like C<getsockname>.  If all you want is to read or write to the socket,
+you should use C<stdin_handle> and C<stdout_handle> to get the in and out filehandles
+explicitly.
+
+=cut
+
+sub stdio_handle {
+    my $self = shift;
+    $self->{'_stdio_handle'} = shift if (@_);
+    return $self->{'_stdio_handle'};
+}
+
+=head2 stdin_handle
+
+Returns a filehandle used for input from the client.  By default, 
+returns whatever was set with C<stdio_handle>, but a subclass
+could do something interesting here (see L<HTTP::Server::Simple::Logger>).
+
+=cut
+
+sub stdin_handle {
+    my $self = shift;
+    return $self->stdio_handle;
+} 
+
+=head2 stdout_handle
+
+Returns a filehandle used for output to the client.  By default, 
+returns whatever was set with C<stdio_handle>, but a subclass
+could do something interesting here (see L<HTTP::Server::Simple::Logger>).
+
+=cut
+
+sub stdout_handle {
+    my $self = shift;
+    return $self->stdio_handle;
+} 
+
+
 =head1 IMPORTANT SUB-CLASS METHODS
 
 A selection of these methods should be provided by sub-classes of this
@@ -248,14 +368,6 @@ of keys of this list.
   peername     Remote name           "200.2.4.5", "foo.com"
   peeraddr     Remote address        "200.2.4.5", "::1"
   localname    Local interface       "localhost", "myhost.com"
-  headers      All HTTP/1.0+ headers  (see below)
-
-The C<headers> method contains I<all> HTTP headers, in order, without
-being parsed.  It is up to the sub-class to either grok them or ignore
-them.
-
-Note that if the client is a HTTP/0.9 client, then there will be no
-headers at all.
 
 =cut
 
@@ -285,9 +397,11 @@ C<headers()> in your sub-class and receive them all at once.
 
 Finally, you can define handlers to receive individual HTTP headers.
 This can be useful for very simple SOAP servers (to name a
-crack-fueled standard that defines its own special HTTP headers).  Eg,
-for a header called C<Content-Length>, you would define the method
-C<content_length()> in your sub-class.
+crack-fueled standard that defines its own special HTTP headers). 
+
+To do so, you'll want to define the C<header()> method in your subclass.
+That method will be handed a (key,value) pair of the header name and the value.
+
 
 =cut
 
@@ -299,14 +413,6 @@ sub headers {
     while ( my ($header, $value) = splice @$headers, 0, 2 ) {
 	if ( $can_header ) {
 	    $self->header($header => $value)
-	} else {
-	    (my $method = lc($header)) =~ s{-}{_}g;
-
-	    # FIXME - security - this is probably very dangerous
-	    # and probably OTT
-	    $self->$method($value)
-		if !defined &$method  # stop really dumb stuff
-		    and $self->can($method);
 	}
     }
 }
@@ -344,6 +450,7 @@ sub print_banner {
 
 Parse the HTTP request line.
 
+Returns three values, the request method, request URI and the protocol
 Sub-classed versions of this should return three values - request
 method, request URI and proto
 
@@ -351,15 +458,25 @@ method, request URI and proto
 
 sub parse_request {
     my $self = shift;
-    defined($_ = <STDIN>)
-	or return undef;
-    chomp;
-    m/^(\w+)\s+(\S+)(?:\s+(\S+))?\r?$/
+    my $chunk;
+    while ( sysread( STDIN, my $buff, 1 ) ) {
+        last if $buff eq "\n";
+        $chunk .= $buff;
+    }
+    defined($chunk) or return undef;
+    $_ = $chunk;
+
+    m/^(\w+)\s+(\S+)(?:\s+(\S+))?\r?$/;
+    my $method = $1 || '';
+    my $uri = $2 || '';
+    my $protocol = $3 || '';
+
+    return($method, $uri, $protocol);
 }
 
 =head2 parse_headers
 
-Parse extra RFC822-style headers with the request.
+Parse incoming HTTP headers from STDIN.
 
 Remember, this is a B<simple> HTTP server, so nothing intelligent is
 done with them C<:-)>.
@@ -374,14 +491,20 @@ sub parse_headers {
 
     my @headers;
 
-    while (<STDIN>) {
-        s/[\r\l\n\s]+$//;
-        if (/^([\w\-]+): (.+)/i) {
-	    push @headers, $1 => $2;
+    my $chunk = '';
+    while ( sysread( STDIN, my $buff, 1 ) ) {
+        if ( $buff eq "\n" ) {
+            $chunk =~ s/[\r\l\n\s]+$//;
+            if ( $chunk =~ /^([\w\-]+): (.+)/i ) {
+                push @headers, $1 => $2;
+            }
+            last if ( $chunk =~ /^$/ );
+            $chunk = '';
         }
-        last if (/^$/);
+        else { $chunk .= $buff }
     }
-    \@headers;
+
+    return(\@headers);
 }
 
 
@@ -400,9 +523,16 @@ sub setup_listener {
     setsockopt( HTTPDaemon, SOL_SOCKET, SO_REUSEADDR, pack( "l", 1 ) )
       or warn "setsockopt: $!";
     bind( HTTPDaemon,
-	  sockaddr_in( $self->port(), ( $self->host ? inet_aton($self->host)
-					: INADDR_ANY ) ) )
-	or die "bind: $!";
+        sockaddr_in(
+            $self->port(),
+            (
+                $self->host
+                ? inet_aton( $self->host )
+                : INADDR_ANY
+            )
+        )
+      )
+      or die "bind: $!";
     listen( HTTPDaemon, SOMAXCONN ) or die "listen: $!";
 
 }
@@ -426,10 +556,10 @@ sub bad_request {
 
 =head1 AUTHOR
 
-Copyright (c) 2004-2005 Jesse Vincent, jesse@bestpractical.com.
+Copyright (c) 2004-2005 Jesse Vincent, <jesse@bestpractical.com>.
 All rights reserved.
 
-Marcu Ramberg contributed tests, cleanup, etc
+Marcus Ramberg <drave@thefeed.no> contributed tests, cleanup, etc
 
 Sam Vilain, <samv@cpan.org> contributed the CGI.pm split-out and
 header/setup API.
