@@ -1,7 +1,7 @@
 
 package HTTP::Server::Simple::CGI;
 
-use base qw(HTTP::Server::Simple);
+use base qw(HTTP::Server::Simple HTTP::Server::Simple::CGI::Environment);
 use strict;
 use warnings;
 
@@ -9,8 +9,6 @@ use CGI ();
 
 our $VERSION = $HTTP::Server::Simple::VERSION;
 
-my %clean_env=%ENV;
-
 =head1 NAME
 
 HTTP::Server::Simple::CGI - CGI.pm-style version of HTTP::Server::Simple
@@ -30,9 +28,8 @@ start-up state.
 =cut
 
 sub accept_hook {
-    %ENV= ( %clean_env,
-	    SERVER_SOFTWARE => "HTTP::Server::Simple/$VERSION"
-	  );
+    my $self = shift;
+    $self->setup_environment(@_);
 }
 
 =head2 post_setup_hook
@@ -41,72 +38,26 @@ sub accept_hook {
 
 =cut
 
-sub post_setup_hook{
-
-    $ENV{SERVER_URL} ||=
-	("http://".$ENV{SERVER_NAME}.":".$ENV{SERVER_PORT}."/");
-
+sub post_setup_hook {
+    my $self = shift;
+    $self->setup_server_url;
     CGI::initialize_globals();
 }
 
 =head2 setup
 
-This method sets up environment variables based on various
+This method sets up CGI environment variables based on various
 meta-headers, like the protocol, remote host name, request path, etc.
 
-=cut
+See the docs in L<HTTP::Server::Simple> for more detail.
 
-our %env_mapping =
-    ( protocol => "SERVER_PROTOCOL",
-      localport => "SERVER_PORT",
-      localname => "SERVER_NAME",
-      path => "PATH_INFO",
-      request_uri => "REQUEST_URI",
-      method => "REQUEST_METHOD",
-      peeraddr => "REMOTE_ADDR",
-      peername => "REMOTE_HOST",
-      query_string => "QUERY_STRING",
-    );
+=cut
 
 sub setup {
-    no warnings 'uninitialized';
     my $self = shift;
-
-    while ( my ($item, $value) = splice @_, 0, 2 ) {
-	if ( $self->can($item) ) {
-	    $self->$item($value);
-	} elsif ( my $k = $env_mapping{$item} ) {
-	    $ENV{$k} = $value;
-	}
-    }
-
+    $self->setup_environment_from_metadata(@_);
 }
 
-=head2  headers
-
-This method sets up the process environment in icky-CGI style based on
-the HTTP input headers.
-
-=cut
-
-sub headers {
-    my $self = shift;
-    my $headers = shift;
-
-    while ( my ($tag, $value) = splice @_, 0, 2 ) {
-	$tag = uc($tag);
-	$tag =~ s/^COOKIES$/COOKIE/;
-	$tag =~ s/-/_/g;
-	$tag = "HTTP_" . $tag
-	    unless $tag =~ m/^CONTENT_(?:LENGTH|TYPE)$/;
-
-	if ( exists $ENV{$tag} ) {
-	    $ENV{$tag} .= "; $value";
-	} else {
-	    $ENV{$tag} = $value;
-	}
-    }
-}
 
 =head2 handle_request CGI
 
@@ -153,11 +104,11 @@ __DATA__
   <body>
     <h1>Congratulations!</h1>
 
-    <p>You now have a functional HTTP::Server::Simple running.
+    <p>You now have a functional HTTP::Server::Simple::CGI running.
       </p>
 
     <p><i>(If you're seeing this page, it means you haven't subclassed
-      HTTP::Server::Simple, which you'll need to do to make it
+      HTTP::Server::Simple::CGI, which you'll need to do to make it
       useful.)</i>
       </p>
   </body>
